import torch
import math
from typing import Callable

# We can use the actual lipschitz constant with respect with just the max
# singular value  of A.T A and remove the line search.

def _line_search(x, lambda_init, f, grad_f, prox_g):
    """ https://stanford.edu/~boyd/papers/pdf/prox_slides.pdf; Slide 20 """
    beta = 0.5
    lambda_ = lambda_init
    grad_fx = grad_f(x)
    f_x = f(x)
    while True:
        z = prox_g(x - lambda_ * grad_fx, lambda_)
        norm_z_m_x = torch.linalg.vector_norm(z - x)
        if f(z) <= f_x + grad_fx @ (z - x) + 1/2 / lambda_ * norm_z_m_x**2:
            break
        lambda_ = beta * lambda_

    return lambda_, z, norm_z_m_x

def prox_grad_w_linesearch(f: Callable[[torch.Tensor], torch.Tensor],
              grad_f: Callable[[torch.Tensor], torch.Tensor],
              prox_g: Callable[[torch.Tensor], torch.Tensor],
              x0: torch.Tensor, _, max_iters=5000, eps=1e-3) -> tuple[torch.Tensor, int]:
    x = x0
    norm_delta_x = float('inf')
    lambda_ = 1.0
    k = 0
    epsilon = eps * math.sqrt(x0.numel())

    while norm_delta_x > epsilon and k < max_iters:
        lambda_, x, norm_delta_x = _line_search(x, lambda_, f, grad_f, prox_g)
        k += 1

    return x, k

def accel_prox_grad_w_linesearch(f: Callable[[torch.Tensor], torch.Tensor],
                    grad_f: Callable[[torch.Tensor], torch.Tensor],
                    prox_g: Callable[[torch.Tensor], torch.Tensor],
                    x0: torch.Tensor, _, max_iters=5000, eps=1e-3) -> tuple[torch.Tensor, int]:
    xprev = 0
    x = x0
    norm_delta_x = float('inf')
    lambda_ = 1.0
    k = 0
    epsilon = eps * math.sqrt(x0.numel())

    while norm_delta_x > epsilon and k < max_iters:
        y = x + k / (k + 3) * (x - xprev)
        xprev = x
        lambda_, x, _ = _line_search(y, lambda_, f, grad_f, prox_g)
        norm_delta_x = torch.linalg.vector_norm(x - xprev).item()
        k += 1

    return x, k

def prox_grad(_,
              grad_f: Callable[[torch.Tensor], torch.Tensor],
              prox_g: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
              x0: torch.Tensor,
              lipschitz_constant: torch.Tensor,
              max_iters=5000, eps=1e-3) -> tuple[torch.Tensor, int]:
    x = x0
    lambda_ = 0.8 / lipschitz_constant
    k = 0
    grad_fx = grad_f(x)
    err = float('inf')
    epsilon = eps * math.sqrt(x0.numel())

    while err > epsilon and k < max_iters:
        #lambda_, x, norm_delta_x = _line_search(x, lambda_, f, grad_f, prox_g)
        intermediate_x = x - lambda_ * grad_fx
        x = prox_g(intermediate_x, lambda_)
        k += 1
        grad_fx = grad_f(x)
        err = torch.linalg.vector_norm(
                grad_fx - 1 / lambda_ * (x - intermediate_x)
            )
        #print(f"{k=}, {err=}, {lambda_=}")

    return x, k

def accel_prox_grad(_,
                    grad_f: Callable[[torch.Tensor], torch.Tensor],
                    prox_g: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
                    x0: torch.Tensor,
                    lipschitz_constant: torch.Tensor,
                    max_iters=5000, eps=1e-3) -> tuple[torch.Tensor, int]:
    xprev = 0
    x = x0
    err = float('inf')
    lambda_ = 0.8 / lipschitz_constant
    k = 0
    epsilon = eps * math.sqrt(x0.numel())

    while err > epsilon and k < max_iters:
        y = x + k / (k + 3) * (x - xprev)
        xprev = x

        intermediate_y = y - lambda_ * grad_f(y)
        x = prox_g(intermediate_y, lambda_)
        k += 1
        err = torch.linalg.vector_norm(
                grad_f(x) - 1 / lambda_ * (x - intermediate_y)
            )
        #print(f"{k=}, {err=}, {lambda_=}")

    return x, k



import torch
import math

from linops import aslinearoperator


def admm(A, b, prox_R, x0, device, debug=False, max_iters=1000, eps_abs=1e-6,
        eps_rel=1e-3):
    """
        Solves problem of the form
            argmin_x R(x) + (1/2)||b - A x||_2^2

        prox_R is the proximal operator of R.
    """
    A = aslinearoperator(A)
    ATb = A.T @ b

    def prox_half_l2_norm_squared(v, lambda_):
        """
        (1/2)||A x - b||_2^2 = (1/2)(A x - b)^T (A x - b) =
            (Ax)^T Ax - (Ax)^T b - b^T Ax + b^T b =
            (1/2) x^T (A^T A) x - b^T A x + (1/2)||b||_2^2
    
        See \\S 6.1.1 of prox_algs.pdf to see that 
        prox_{lambda_ f}(v) = (I + lambda_ A^T A)^-1 (v + lambda_ A^T b)
        """
        RHS = v + lambda_ * ATb
        return A.solve_I_p_lambda_AT_A_x_eq_b(lambda_, RHS)

    problem_size = x0.shape.numel()
    rho = problem_size**1.5
    x = x0
    z = x0
    u = torch.zeros_like(x0, device=device)
    iters = 0
    if debug:
        err_pri = []
        err_dual = []
    r_pri_norm = float('inf')
    r_dual_norm = float('inf')
    x_norm = torch.linalg.vector_norm(x).item()
    z_norm = torch.linalg.vector_norm(z).item()
    u_norm = 0.0

    scaled_eps_abs =  eps_abs * math.sqrt(problem_size)
    mu = 10

    # See \S4.4 of prox_algs.pdf
    while (iters < max_iters and
            (r_pri_norm > scaled_eps_abs + eps_rel * max(x_norm, z_norm)
                or r_dual_norm > scaled_eps_abs + eps_rel * rho * u_norm)):
        if debug:
            print(f"{iters=} {x_norm=} {z_norm=} {u_norm=}")
        x = prox_R(z - u, 1/rho)
        x_norm = torch.linalg.vector_norm(x).item()
        z_new = prox_half_l2_norm_squared(x + u, 1/rho)
        r_dual_norm = rho * torch.linalg.vector_norm(z_new - z).item()
        z = z_new
        z_norm = torch.linalg.vector_norm(z).item()
        u = u + x - z
        u_norm = torch.linalg.vector_norm(u).item()
        iters += 1
        r_pri_norm = torch.linalg.vector_norm(x - z).item()

        if r_pri_norm > mu * r_dual_norm:
            if debug:
                print(f"{iters=} upping rho, {r_pri_norm=}, {r_dual_norm=}, {rho=}")
            rho = rho * 2
            u = u / 2
        elif r_dual_norm > mu * r_pri_norm:
            if debug:
                print(f"{iters=} downing rho, {r_pri_norm=}, {r_dual_norm=}, {rho=}")
            rho = rho / 2
            u = u * 2
    if debug:
        print(f"Final ADMM {iters=}")
    return x, z, u, iters
